// SRP Server Library
// srpserver.c
// Created: 12/3/99
// Modified: 12/3/99

// 12-3  mpiecuch     Initial Creation 

// Compile: gcc -lpthread srpserver.c


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>
/* #include "srp_stats.h" */
#include "srp.h"


struct srp_server_socket *srp_server_start(int sock, unsigned int flags, const struct sockaddr *to, int tolen) {

  int i;
  struct srp_server_socket *srp_sock;
  struct srp_buffer_unit *srp_buff;

  srp_sock = malloc(sizeof(struct srp_server_socket));

  if (srp_sock == NULL) {
    printf("Could not allocate memory for srp socket");
    return(NULL);
  }


  srp_buff = malloc(sizeof(struct srp_buffer_unit [SRP_BUFFER_SIZE]));

  if (srp_buff == NULL) {
    printf("Could not allocate memory for srp buffer of size %ld", SRP_BUFFER_SIZE);
    return(NULL);
  }

  srp_sock->srp_buffer = srp_buff;

  for(i=0;i<SRP_BUFFER_SIZE;i++) {
    srp_sock->srp_buffer[i].size = 0;
    srp_sock->srp_buffer[i].data_id = 0;
    srp_sock->srp_buffer[i].data = NULL;
  }

  srp_sock->socket = sock;
  srp_sock->srp_cur_data_id = 0;
  srp_sock->flags = flags;
  memcpy(&(srp_sock->to), to, sizeof(struct sockaddr));
  srp_sock->tolen = tolen;

  pthread_mutex_init(&(srp_sock->mutex), NULL);

  pthread_create(&(srp_sock->srp_recv_thread), NULL, (void *) &srp_recv_messages, srp_sock);

  return(srp_sock);

}


int srp_sendto(struct srp_server_socket *srp_sock, const void *msg, int len) {

  int buffer_pos;
  int return_val = 0;
  struct srp_data_packet pkt;
  void *mem_buff;

  if (len > SRP_MAX_DATA_SIZE) {
    // TO DO: Add better error handling here
    printf("%d bytes of data is too large for transmission", len);
    return(ESRPSIZEERROR);
  }

  // Calculate buffer position
  buffer_pos = srp_sock->srp_cur_data_id % SRP_BUFFER_SIZE;



  // Allocate memory for new data
  mem_buff = malloc(len);
  if (mem_buff == NULL) {
    // TO DO: Add better error handling here 
    printf("Cannot allocate %d bytes of buffer space", len);
    return(ENOMEM);
  }
  memcpy(mem_buff, msg, len);

  // Lock so that the buffer can be changed safely
  if (pthread_mutex_lock(&(srp_sock->mutex)) < 0)
   {
    return (EINVAL);
   } 
  // Free any memory that was allocated to previous data.
  free(srp_sock->srp_buffer[buffer_pos].data);

  // Fill buffer with new message
  srp_sock->srp_buffer[buffer_pos].size = len;
  srp_sock->srp_buffer[buffer_pos].data_id = srp_sock->srp_cur_data_id;
  srp_sock->srp_buffer[buffer_pos].data = mem_buff;

  // Threads are safe again
  if (pthread_mutex_unlock(&(srp_sock->mutex)) < 0)
    {
     return(EINVAL);
    }


  // Format SRP packet
  pkt.srp_head.srp_type = SRP_TYPE_DATA;
  pkt.srp_head.timestamp = SRP_NO_TIME;
  pkt.srp_data_head.srp_data_id = srp_sock->srp_cur_data_id;
  pkt.srp_data_head.srp_data_size = len;
  memcpy (pkt.data, msg, len);

  printf("Sending data id: %d size: %d\n", pkt.srp_data_head.srp_data_id, pkt.srp_data_head.srp_data_size);

  // Send SRP packet
  return_val = sendto(srp_sock->socket, &pkt, sizeof(struct srp_data_packet_header) + sizeof(struct srp_packet_header)+ len, srp_sock->flags, &(srp_sock->to), srp_sock->tolen);


#ifdef SRP_STAT
  stat.data_pkts_sent++;
  stat.data_bytes_sent += sizeof(struct srp_data_packet_header) + sizeof(struct srp_packet_header) + len;
#endif

  // Increment cur_data_id
 srp_sock->srp_cur_data_id = (srp_sock->srp_cur_data_id + 1) % SRP_MAX_DATA_ID;


  return(return_val);


}



int srp_recv_messages(void *sock_ptr) {


  int data_id, buffer_pos, buffer_hit_flag;
  struct srp_server_socket *srp_sock;
  struct srp_nack_packet nack;
  struct srp_data_packet pkt;
  struct srp_probe_packet probe;


  if (sock_ptr == NULL) {
      printf("Srp socket not initilized");
      return(ESRPNULLPTR);
  }

  srp_sock = (struct srp_server_socket *) sock_ptr;


  while(1) {

    // Receive packet from socket
    // Future work: May want to check where packets come from.
    if (recvfrom(srp_sock->socket, &nack, sizeof(struct srp_nack_packet), 0, NULL, NULL) < 0) {
      // Recv error (socket closed?), stop thread I guess
      printf("Could not receive from socket\n");
      pthread_exit(NULL);
    }

    if (nack.srp_head.srp_type == SRP_TYPE_PROBE) {
      
      printf("Time Probe received: time: %ld\n", nack.srp_head.timestamp);

      probe.srp_head.srp_type = SRP_TYPE_PROBE;
      probe.srp_head.timestamp = nack.srp_head.timestamp;

      if (sendto(srp_sock->socket, &probe, sizeof(struct srp_probe_packet), 0, &(srp_sock->to), srp_sock->tolen) < 0) {
	// TO DO: Provide better error handling here
	printf("Could not time probe");
	return(EFAULT);	  
      }    


    } 

    if (nack.srp_head.srp_type == SRP_TYPE_NACK) {

      printf("Nack received: status: %d, data_id: %d\n", nack.status, nack.srp_data_id);

      // Interpret message
      if (nack.status == SRP_STATUS_MISSING) {
	data_id = nack.srp_data_id;
	

	// Look for missing data in buffer
	
	// Calculate buffer position
	buffer_pos = data_id % SRP_BUFFER_SIZE;   
	

	// Lock so that shared buffer data can be read safely
	if ( pthread_mutex_lock(&(srp_sock->mutex)) < 0)
	  {
	    return(EINVAL);
	  }

	if (srp_sock->srp_buffer[buffer_pos].data_id == data_id) {
	  // Buffer hit
	  printf("Buffer Hit\n");
	  buffer_hit_flag = 1;
	
	  // Get Buffered data
	  
	  // Format SRP packet
	  pkt.srp_head.srp_type = SRP_TYPE_DATA;
	  pkt.srp_head.timestamp = nack.srp_head.timestamp;
	  pkt.srp_data_head.srp_data_id = srp_sock->srp_buffer[buffer_pos].data_id;
	  pkt.srp_data_head.srp_data_size = srp_sock->srp_buffer[buffer_pos].size;
	  memcpy(pkt.data, srp_sock->srp_buffer[buffer_pos].data, srp_sock->srp_buffer[buffer_pos].size);

	} else {
	  // Buffer Miss
	  printf("Buffer Miss\n");
	  buffer_hit_flag = 0;
#ifdef SRP_STAT
	  stat.buffer_misses++;
#endif
	}
     
	// Threads are safe again
	if (pthread_mutex_unlock(&(srp_sock->mutex)) < 0)
	  {
	    return(EINVAL);
	  }
     
	if (buffer_hit_flag) {

	  // Send SRP packet
	  // NOTE: may want to do something to the flags parameter of the send command
	  printf("Sending retransmission: data_id: %d buffer_pos: %d\n", data_id, buffer_pos);
	  if (sendto(srp_sock->socket, &pkt, sizeof(struct srp_data_packet_header) + sizeof(struct srp_packet_header)+ pkt.srp_data_head.srp_data_size, 0, &(srp_sock->to), srp_sock->tolen) < 0) {
	    // TO DO: Provide better error handling here
	    printf("Could not send retransmission");
	    return(EFAULT);	  
	  }  

#ifdef SRP_STAT
	  stat.retrans_sent++;
	  stat.retrans_bytes_sent += sizeof(struct srp_data_packet_header) + sizeof(struct srp_packet_header)+ pkt.srp_data_head.srp_data_size;
#endif


	}

      }

    }

  }


}

int srp_server_stop(struct srp_server_socket *srp_sock) {
  
  int i;

  // Stop the Receiving thread
  if (pthread_cancel(srp_sock->srp_recv_thread)) {
    printf("Could not cancel receiving thread\n");
   return(ESRCH);
  }

  // Clear the data out of the buffer
  for(i=0;i<=SRP_BUFFER_SIZE;i++) {
    if (srp_sock->srp_buffer[i].data != NULL) {
      free(srp_sock->srp_buffer[i].data);
    } 
  }
   
  // free the data buffer
  free(srp_sock->srp_buffer);


  // free the data from the srp socket
  free(srp_sock);


}










