// SRP Server
// srpserver.c
// Created: 11/11/99
// Modified: 11/21/99

// 11-11  mpiecuch     Initial Creation 
// 11-20  mpiecuch     Added semaphore and thread cancelation
// 11-21  mpiecuch     Added Statistics

// Compile: gcc -lpthread srpserver.c


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>
#include "srp_stats.h"


// This eventually goes into srp.h

#define SRP_STATUS_MISSING 0
#define SRP_MAX_DATA_SIZE 1480


#ifndef SRP_STAT

#define SRP_BUFFER_SIZE 10
#define SRP_MAX_DATA_ID 255
#define FRAME_SPEED 5
#define NUM_FRAMES 10

#endif


struct srp_buffer_unit {
  int size;
  unsigned int data_id;
  void *data;
};

struct srp_data_packet_header {
  unsigned int srp_data_id;
  unsigned int srp_data_size;
};

struct srp_data_packet {
  struct srp_data_packet_header srp_head;
  char data[SRP_MAX_DATA_SIZE];
};

struct srp_nack_packet {
  char status;
  unsigned int srp_data_id;
};

struct srp_socket {
  int socket;
  unsigned int flags;
  struct sockaddr to; 
  int tolen;
  struct srp_buffer_unit *srp_buffer;
  int srp_cur_data_id;
  pthread_t srp_recv_thread;
  pthread_mutex_t mutex;
};


struct srp_socket *srp_server_start(int, unsigned int, const struct sockaddr *, int);
int srp_sendto(struct srp_socket *, const void *, int);
void srp_recv_messages(void *);
void srp_print_buffer(struct srp_socket *);
void srp_server_stop(struct srp_socket *);

// End of srp.h info


// Info for testing code

#define MAX_BACKLOG 10


// Main defined to test srp code
int main(int argc, char *argv) {

  char test_data[255];
  int test_int;
  struct srp_socket *srp_sock;
  int udp_sock;
  struct sockaddr_in localaddr, peeraddr;
  int size_localaddr, port;
  char host[50];
  struct srp_nack_packet srp_ready_msg;
  struct timeval tv;


#ifdef SRP_STAT
  printf("Statistics Version\n");
  srp_stat_init(&stat);
  srp_stat_server_read("server_params.txt");
#endif

  tv.tv_sec = 0;
  tv.tv_usec = FRAME_SPEED;


  // Create normal UDP socket
  if ((udp_sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {perror("Socket Error! "); exit(1);}

  // Setup necessary UDP socket stuff
  localaddr.sin_family = AF_INET;
  localaddr.sin_port = 0;
  localaddr.sin_addr.s_addr = INADDR_ANY;
  size_localaddr = sizeof(localaddr);


  // Bind socket and get host/port info
  if ((bind(udp_sock, (struct sockaddr *) &localaddr, sizeof(localaddr))) < 0) {perror("Bind Error! "); exit(1);}
  if ((getsockname(udp_sock, (struct sockaddr *)  &localaddr, &size_localaddr )) < 0) {perror("Get Port Error! "); exit(1);}
  port = localaddr.sin_port;
  if ((gethostname(host, sizeof(host))) < 0) {perror("Get Host Error! "); exit(1);}

  printf("host: %s, port: %d\n", host, ntohs(port));


  // Get a ready (nack right now) message from the client to get the address to send to.
  recvfrom(udp_sock, &srp_ready_msg, sizeof(struct srp_nack_packet), 0, (struct sockaddr *) &peeraddr, &size_localaddr);

  printf("Initializing srp with: %s\n", inet_ntoa(peeraddr.sin_addr));


  // Create SRP socket and start retransmission listening thread
  srp_sock = srp_server_start(udp_sock, 0, (struct sockaddr *) &peeraddr, sizeof(struct sockaddr));



  // Keep sending test data to client every 5 seconds
  test_int = 0;
  while(test_int<=NUM_FRAMES) {
    printf("Sending frame: %d\n", test_int);
    // Note: \0 is to make sure test client can print data as string
    sprintf(test_data, "Test data: %d\0", test_int++);
    // Note + 1 is to make sure \0 is received at the test client
    srp_sendto(srp_sock, test_data, strlen(test_data) + 1);
    //srp_print_buffer(srp_sock);
    select(0, NULL, NULL, NULL, &tv);
    tv.tv_sec = 0;
    tv.tv_usec = FRAME_SPEED;
  }
  
  srp_server_stop(srp_sock);

#ifdef SRP_STAT
  srp_stat_print(stat);
  srp_stat_server_write(stat, "server_stats.txt");
#endif


}

struct srp_socket *srp_server_start(int sock, unsigned int flags, const struct sockaddr *to, int tolen) {

  int i;
  struct srp_socket *srp_sock;
  struct srp_buffer_unit *srp_buff;

  srp_sock = malloc(sizeof(struct srp_socket));

  if (srp_sock == NULL) {
    // TO DO: Add better error handling
    printf("Could not allocate memory for srp socket");
    return(NULL);
  }


  srp_buff = malloc(sizeof(struct srp_buffer_unit [SRP_BUFFER_SIZE]));

  if (srp_sock == NULL) {
    // TO DO: Add better error handling
    printf("Could not allocate memory for srp buffer of size %ld", SRP_BUFFER_SIZE);
    return(NULL);
  }

  srp_sock->srp_buffer = srp_buff;

  for(i=0;i<SRP_BUFFER_SIZE;i++) {
    srp_sock->srp_buffer[i].size = 0;
    srp_sock->srp_buffer[i].data_id = 0;
    srp_sock->srp_buffer[i].data = NULL;
  }

  srp_sock->socket = sock;
  srp_sock->srp_cur_data_id = 0;
  srp_sock->flags = flags;
  memcpy(&(srp_sock->to), to, sizeof(struct sockaddr));
  srp_sock->tolen = tolen;

  pthread_mutex_init(&(srp_sock->mutex), NULL);

  pthread_create(&(srp_sock->srp_recv_thread), NULL, (void *) &srp_recv_messages, srp_sock);

  return(srp_sock);

}


int srp_sendto(struct srp_socket *srp_sock, const void *msg, int len) {

  int buffer_pos;
  int return_val = 0;
  struct srp_data_packet pkt;
  void *mem_buff;


  printf("Enter data id: ");
  scanf("%ld", &(srp_sock->srp_cur_data_id));


  if (len > SRP_MAX_DATA_SIZE) {
    // TO DO: Add better error handling here
    printf("%d bytes of data is too large for transmision", len);
    exit(1);
  }

  // Calculate buffer position
  buffer_pos = srp_sock->srp_cur_data_id % SRP_BUFFER_SIZE;



  // Allocate memory for new data
  mem_buff = malloc(len);
  if (mem_buff == NULL) {
    // TO DO: Add better error handling here 
    printf("Cannot allocate %d bytes of buffer space", len);
    exit(1);
  }
  memcpy(mem_buff, msg, len);

  // Lock so that the buffer can be changed safely
  pthread_mutex_lock(&(srp_sock->mutex));

  // Free any memory that was allocated to previous data.
  free(srp_sock->srp_buffer[buffer_pos].data);

  // Fill buffer with new message
  srp_sock->srp_buffer[buffer_pos].size = len;
  srp_sock->srp_buffer[buffer_pos].data_id = srp_sock->srp_cur_data_id;
  srp_sock->srp_buffer[buffer_pos].data = mem_buff;

  // Threads are safe again
  pthread_mutex_unlock(&(srp_sock->mutex));


  // Format SRP packet
  pkt.srp_head.srp_data_id = srp_sock->srp_cur_data_id;
  pkt.srp_head.srp_data_size = len;
  memcpy (pkt.data, msg, len);


  // Send SRP packet
  return_val = sendto(srp_sock->socket, &pkt, sizeof(struct srp_data_packet_header) + len, srp_sock->flags, &(srp_sock->to), srp_sock->tolen);

#ifdef SRP_STAT
  stat.data_pkts_sent++;
  stat.data_bytes_sent += sizeof(struct srp_data_packet_header) + len;
#endif

  // Increment cur_data_id
  srp_sock->srp_cur_data_id = (srp_sock->srp_cur_data_id + 1) % SRP_MAX_DATA_ID;


  return(return_val);


}

void srp_print_buffer(struct srp_socket *srp_sock) {

  int i, j;
  char *temp;

  for(i=0;i<SRP_BUFFER_SIZE;i++) {
    printf("Buffer Slot %d:\n", i);
    if (srp_sock->srp_buffer[i].size > 0) {
      temp = (char *) srp_sock->srp_buffer[i].data;
      for (j=0;j<srp_sock->srp_buffer[i].size;j++) {
	printf("%c", temp[j]);
      }
    }
    printf("\n----------------\n");
  }
}

void srp_recv_messages(void *sock_ptr) {


  int data_id, buffer_pos, buffer_hit_flag;
  struct srp_socket *srp_sock;
  struct srp_nack_packet nack;
  struct srp_data_packet pkt;
  fd_set readset, writeset, exceptset;


  if (sock_ptr == NULL) {
      // TO DO: Provide better error handling here
      printf("Srp socket not initilized");
  }

  srp_sock = (struct srp_socket *) sock_ptr;


  while(1) {

    // Receive packet from socket
    // Future work: May want to check where packets come from.
    if (recvfrom(srp_sock->socket, &nack, sizeof(struct srp_nack_packet), 0, NULL, NULL) < 0) {
      // Recv error (socket closed?), stop thread I guess
      printf("Could not receive from socket\n");
      pthread_exit(NULL);
    }

    

    printf("Nack received: status: %d, data_id: %d\n", nack.status, nack.srp_data_id);

    // Interpret message
    if (nack.status == SRP_STATUS_MISSING) {
      data_id = nack.srp_data_id;


      // Look for missing data in buffer

      // Calculate buffer position
      buffer_pos = data_id % SRP_BUFFER_SIZE;   


      // Lock so that shared buffer data can be read safely
      pthread_mutex_lock(&(srp_sock->mutex));

      if (srp_sock->srp_buffer[buffer_pos].data_id == data_id) {
	// Buffer hit
	printf("Buffer Hit\n");
	buffer_hit_flag = 1;
	
	// Get Buffered data

	// Format SRP packet
	pkt.srp_head.srp_data_id = srp_sock->srp_buffer[buffer_pos].data_id;
	pkt.srp_head.srp_data_size = srp_sock->srp_buffer[buffer_pos].size;
	memcpy(pkt.data, srp_sock->srp_buffer[buffer_pos].data, srp_sock->srp_buffer[buffer_pos].size);

      } else {
	// Buffer Miss
	printf("Buffer Miss\n");
	buffer_hit_flag = 0;
#ifdef SRP_STAT
	stat.buffer_misses++;
#endif
      }
     
      // Threads are safe again
      pthread_mutex_unlock(&(srp_sock->mutex));
     

      if (buffer_hit_flag) {

	// Send SRP packet
	// NOTE: may want to do something to the flags parameter of the send command
	printf("Sending retransmission: data_id: %d buffer_pos: %d\n", data_id, buffer_pos);
	if (sendto(srp_sock->socket, &pkt, sizeof(struct srp_data_packet_header) + pkt.srp_head.srp_data_size, 0, &(srp_sock->to), srp_sock->tolen) < 0) {
	  // TO DO: Provide better error handling here
	  printf("Could not send retransmission");	  
	}  

#ifdef SRP_STAT
	stat.retrans_sent++;
	stat.retrans_bytes_sent += sizeof(struct srp_data_packet_header) + pkt.srp_head.srp_data_size;
#endif


      }

    }

  }


}

void srp_server_stop(struct srp_socket *srp_sock) {
  
  int i;

  // Stop the Receiving thread
  if (pthread_cancel(srp_sock->srp_recv_thread)) {
    printf("Could not cancel receiving thread\n");
  }

  // Clear the data out of the buffer
  for(i=0;i<=SRP_BUFFER_SIZE;i++) {
    if (srp_sock->srp_buffer[i].data != NULL) {
      free(srp_sock->srp_buffer[i].data);
    } 
  }
   
  // free the data buffer
  free(srp_sock->srp_buffer);


  // free the data from the srp socket
  free(srp_sock);


}





